# To generate prompts

from .config import DEFAULT_FILTERING_CRITERIA
from .utils import remove_excess


def gen_prompt_filter(criteria:str=DEFAULT_FILTERING_CRITERIA) -> str:
    """
    Generates the system prompt / role to filter content based on a list of criteria.
    """
    return remove_excess(f"""
        You are a rigorous analyst. 
        The user will provide you with content extracted from a website. You must analyze this content and return True if the majority of the Criteria are met.
        ONLY use the content provided and follow strictly the below Instructions:

        ### Instructions: 
        1. Take your time and think step by step.
        2. Carefully go through the below Criteria: {criteria}
        3. If the majority of the Criteria are respected, return the word True and nothing else.
        4. If the Criteria are NOT met at ALL, return the word False followed with a justification where you explain WHY the Criteria are not met at all. Be specific and say which criteria is not met and why.

        ### Important considerations
        - Follow strictly the Instructions. 
        - Return True if the majority of the Criteria are respected. In this case, do not provide any justification. Just return True.
        - If a large majority of the Criteria are STRICTLY NOT met, return the word False followed with the explanation. In this case, you MUST explain which Criteria is not met and WHY.
        - Do not preface your answer by anything.
        """)

def gen_prompt_result(json_output:str, json_example:str=None, additional_consideration:str=None) -> str:
    """
    Provide a list of criteria
    """
    additional_consideration = "" if not additional_consideration else "Very important: You MUST the below considerations in mind when performing the task:\n" + additional_consideration
    example = "" if not json_example else f"""\n### Example of output:\n{json_example}\n"""
    return remove_excess(f"""
        You are a rigorous analyst. 
        The user will provide you with content extracted from a website. You must analyze this content and return a JSON object with the information filled when available.
        ONLY use the content available, NEVER INVENT anything and follow strictly the below Instructions:

        ### Instructions: 
        1. Take your time and think step by step.
        2. Review and keep in mind ALL the fields of the below json structure: {json_output}
        3. Go through the content provided by the user and try your best to fill the json accordingly. Take your time. Review the content several times.
        4. If the information is NOT in the content AT ALL, put NA.
        {example}
        ### Important considerations:
        - Follow strictly the Instructions. 
        - Remember that you must construct a JSON object with the provided keys, populated with the relevant information. If an information is not available at ALL or does not apply, use 'NA' for those fields.
        - Return the JSON and the JSON only. Do not preface by anything not even ```json before. 
        - Do your best, my job depends on the quality of your output.
        {additional_consideration}
        """)

def gen_role_summarizer() -> str:
    """
    Quick prompt for summarization. 
    """
    return remove_excess("""
    You are an expert at making summarizations that preserve ALL useful information.
    You take the text submitted by the user and return a detailed summary.
    You MUST ensure that you keep ALL relevant information about the main topic of the content.
                        
    ### Important Considerations: 
    - Do not include anything that is not in the provided text. ALWAYS be truthfull.
    - Do not include any recommendation. IMPORTANT, do not add useless suggestions like 'individuals can subscribe to their monthly newsletter'.
    - Do not preface the summary with anything. Do not put a title. Only return the summary and nothing else.
    - Ensure you keep ALL contact details (email or phone number). Do not consider a website URL as a contact information.
    - Ensure you keep ALL numbers and factual informations.
    Do your best, my job depends on the quality of your work.
    """)

# *************************************************************
if __name__ == "__main__":
    pass


# PIP Package to get specific information from a website using ChatGPT

# First version: 9th of May 2024 by Henry Obegi - Github hanbdle: henryobj
# Updated by XXXXX on XXXXXX
# Contributors: XXXXXXX 

import os

# ***** KEYS 
OAI_KEY = os.getenv("OAI_API_KEY")


# ****** TEXT
OPEN_AI_ISSUE = r"%$144$%" # Arbitrary choice. Just an unlikely string to be output by the LLM. 
ERROR_MESSAGE = "An error occurred and was logged"
WARNING_UNKNOWN = "\033[31mUNKNOWN\033[0m"

# ****** For Web
HTTP_URL_PATTERN = r'^http[s]?://.+'   # Regex pattern to match a URL
HEADERS = {
    "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/58.0.3029.110 Safari/537.3",
}

# ****** Models used for Open AI + Window size
MODEL_CHAT = r"gpt-3.5-turbo-0125"
MODEL_CHAT_STABLE = r"gpt-3.5-turbo"
MODEL_GPT4_TURBO = (
    r"gpt-4-1106-preview"  # Max 128,000 token context window total with 4,096 output
)
MODEL_GPT4_STABLE = r"gpt-4"  # 8K context window and 4,096 output

MODEL_EMB_LARGE = r"text-embedding-3-large"
MODEL_EMB_SMALL = r"text-embedding-3-small"

MAX_TOKEN_OUTPUT_DEFAULT = 300
MAX_TOKEN_GPT4_RESULT = 500
MAX_TOKEN_OUTPUT_DEFAULT_HUGE = 3000
MAX_TOKEN_OUTPUT_FILTER = 10000

MAX_TOKEN_WINDOW_GPT4_TURBO = 128000
MAX_TOKEN_WINDOW_GPT35_TURBO = 16385
MAX_TOKEN_WINDOW_GPT4 = 8192
MAX_TOKEN_OUTPUT_GPT3 = 4096


# ***** Default Prompts and output - can be transformed into objects / pull from a json file in the future.
DEFAULT_FILTERING_CRITERIA = """
### Criteria ###
The VC fund / investor should:
a. Primarily invest in companies at early stages (pre-revenue or very early revenues).
b. Invest in startups with a valuation of approximately $3M to $15M. Check that the fund does NOT typically invest small amounts for disproportionately large equity. For example investing $100K against 10% of equity.
c. Have already invested in Edtech. The VC fund / investor should have experience with SaaS and AI.
d. Not finance companies that are direct competitors in the space of 'training with simulation of interaction with fake clients'.
e. Have made at least 5-10 investments.
f. Be located in the US or the GCC (Gulf Cooperation Council) countries.
### end of the Criteria ###
"""

JSON_OUTPUT_VC = """{
  "Name of the fund": "string",
  "Country": "string",
  "Number of investments": "integer",
  "Size of the fund": "string",
  "One relevant investment to us and why": "string",
  "Who to contact": {
    "Name": "string",
    "Position": "string",
    "Email": "string",
    "Linkedin page": "string"
  },
  "Good reasons to contact them": "string",
  "Useful Infos": "string"
}"""

JSON_EXAMPLE_VC = """{
  "Name of the fund": "Innovate Ventures",
  "Country": "USA",
  "Number of investments": "12",
  "Size of the fund": "$50M",
  "One relevant investment to us and why": "EduTech Innovators - Innovate Ventures invested $200K in 2019. EduTech is relevant because your company does X and EduTech does Y which has synergy with your company on A/B/C.",
  "Who to contact": {
    "Name": "Jane Doe",
    "Position": "Investment Director",
    "Email": "jdoe@innovatevc.com",
    "Linkedin page": "https://www.linkedin.com/in/janedoe"
  },
  "Good reasons to contact them": "Innovate Ventures has a strong focus on Edtech, which aligns with our product. They also have a track record of supporting early-stage startups and can provide valuable mentorship and networks in the US market.",
  "Useful Infos": "The fund recently increased its investment cap for Edtech startups and is actively seeking new innovative projects to finance. It mentioned that X and Y. It is noteworthy that Z."
}"""




#   Main function to scrap the page according to the criteria specified.

from .config import MAX_TOKEN_OUTPUT_DEFAULT_HUGE, MAX_TOKEN_OUTPUT_GPT3, MAX_TOKEN_GPT4_RESULT, MAX_TOKEN_WINDOW_GPT35_TURBO
from .prompts import gen_prompt_filter, gen_prompt_result, gen_role_summarizer
from .web import crawl_website, fetch_content_url, clean_url_to_filename
from .oai import ask_question_gpt, calculate_token, ask_question_gpt4
from .utils import get_now, log_issue


from typing import Optional
import time
import os

def smartscrap(url:str, desired_output:str=None, example_output:str=None, filtering_criteria:str=None, summarization:bool=True, full_website:bool=True, additional_consideration:str=None, verbose:bool=None) -> Optional[str]:
    """
    Main function to scrap a website.

    Args:
        - url (str): The website to scrap
        - desired_output (str): A json like structure with the keys representing the information you want to find
        - filtering_criteria (str): A set of criteria the content must respect for the program to continue
        - summarization (bool): If you want also to output the summary of the website. True by default.
        - additional_consideration (str): If you want to provide additional info to consider when generating the result. Ex: Info about your company.

    Returns:
        - The content or None if issue. The content will also be put in a file named "gptscrapper_url.txt" in the current dir. 

    Note:    
    full_website is True by default. We scrap the full website up to 30 pages by default. Put it to False if you want to scrap ONLY the page.
    Change the crawl_website params to scrap more (or less) pages
    """
    start = time.time()
    try:
        final_content, summary = "", ""
        if full_website:
            if verbose: print(f"👷‍♂️ Crawling the full website {url}. Please be patient...")
            content = str(crawl_website(url))
            
        else:
            if verbose: print(f"👷‍♂️ Crawling the page {url}.")    
            content = fetch_content_url(url)
        if filtering_criteria:
            if verbose: print(f"👷‍♂️ Checking if content matches criteria")
            role_filter = gen_prompt_filter(filtering_criteria)
            buffer_tok = MAX_TOKEN_WINDOW_GPT35_TURBO - calculate_token(role_filter) - calculate_token(content)
            if buffer_tok < MAX_TOKEN_OUTPUT_GPT3-100: # Adding -100 as security
                print(f"The website content is too large for a single prompt - remains only {buffer_tok}\nTBD // TODO for Henry next version\nFor now we will take a subset of the content\n🔴 Information might be missing!")
                safe_removal = int((MAX_TOKEN_OUTPUT_GPT3 - buffer_tok) * 4) * 1.362 # Adding 36% buffer on top
                content = content[:-safe_removal]
            else:
                buffer_tok = max(min(buffer_tok, MAX_TOKEN_OUTPUT_GPT3-100), MAX_TOKEN_OUTPUT_DEFAULT_HUGE) # basically between 3K and 4K
            answer_from_filtergpt = ask_question_gpt(content, role_filter, max_tokens= buffer_tok, verbose=False)
            if not answer_from_filtergpt:
                if verbose: print("Couldn't check the Criteria - END")
                return
            elif "false" in answer_from_filtergpt[:10].lower(): 
                if verbose: print(f"The website is NOT relevant according to the Criteria. Response: {answer_from_filtergpt} - END")
                return
        if summarization:
            if verbose: print(f"👷‍♂️ Generating a summary of the website content")   
            summary = ask_question_gpt(content, gen_role_summarizer(), max_tokens=MAX_TOKEN_OUTPUT_DEFAULT_HUGE,  verbose=False)
        
        if verbose: print(f"👷‍♂️ Using GPT 4 to get the requested data")  
        result = ask_question_gpt4(content, gen_prompt_result(desired_output, example_output, additional_consideration), max_tokens=MAX_TOKEN_GPT4_RESULT,  verbose=False)
        if result:
            title = f"scrapwithgpt_{clean_url_to_filename(url)}.txt"
            if os.path.exists(title):
                title = f"scrapwithgpt_{clean_url_to_filename(url)}_{get_now(True)}.txt" # To avoid overwriting
            if summary:
                final_content = f"### SUMMARY of {url}:\n" + summary.replace(".", ".\n") + "\n\n" 
            final_content += f"### RESULT for {url}:\n" + result
            with open(title, "w") as file:
                file.write(final_content)
            if verbose: print(f"✅ Done - The content is in {title} and was obtained in {round(time.time()-start, 1)} seconds 👌")
        else:
            print("Failed to get the result data - END")
        return final_content
    except Exception as e:
        log_issue(e, smartscrap, f"TBD - add the params here")

# *************************************************************

if __name__ == "__main__":
    pass
    # How to use
    # test_url = "https://www.27v.vc"
    #smartscrap(test_url, desired_output=JSON_OUTPUT_VC, example_output=JSON_EXAMPLE_VC, filtering_criteria=DEFAULT_FILTERING_CRITERIA, verbose=True)
